#include <cassert>
#include <iostream>
#include <cmath>
#include <string>
#include <vector>
#include <iterator>
#include <sstream>
#include <fstream>
#include <EMDecomposer2D.h> 
#include <NHLSOpt.h>

using namespace std; 

extern const string OutputPath ;

EMDecomposer2D::EMDecomposer2D() 
{        
}
                

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//                           Compute Mean by averaging the Max/Min Envelopes
//                                           
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void EMDecomposer2D::ComputeMeanEnvelope(const DataMat& d, const DataMat& MinEnv, const DataMat& MaxEnv, DataMat& Mean)
{        
        int nC(d.NCols()), nR(d.NRows()) ;        
        for (int i(0); i<nR; i++) for(int j(0); j<nC; j++) Mean(i,j) = 0.5 * (MinEnv(i,j) + MaxEnv(i,j)) ; 
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//                                                      Decomposition algorithm
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void EMDecomposer2D::RunDecomposition(const DataMat& data,  int NModes)
{
        int nC(data.NCols()), nR(data.NRows()) ;        
	DataMat d = data ;
        double DR = d.GetDR() ;
        vector<pair<int, int> > Mins, Maxs ;
        d.StationaryPointsStatistical(Mins, Maxs, 1e-8*DR) ;

        DataMat MaxMat(nR, nC) ;
        DataMat MinMat(nR, nC), ExtrMatZero(nR, nC); 
        for (int i(0); i<Mins.size(); i++) MinMat(Mins[i].first, Mins[i].second) = -1 ;
        for (int i(0); i<Maxs.size(); i++) MaxMat(Maxs[i].first, Maxs[i].second) = 1 ;

        DataMat d2 = d ;
	int nm(0) ;

        int KSize(3) ;
        double ScaleIntensity(1.0); 
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////////////// Iterate over modes 
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        for(int l(0); l<NModes; l++) 
        {
                d = d2;
                int nSifts = 2;
		cout << " ------------------------- Mode  " << l+1 << " of " << nSifts <<  " -----------------------------" << endl; 
                ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                for(int n(0); n<nSifts; n++)
                {
                        cout << "\t\t  ------------------------- Sift  " << n+1 << " of " << nSifts  << endl; 
                        ////////////////////////////////////////// Compute Min/Max envelopes 
                        DataMat MinEnv, MaxEnv;                                
                        NHLSMinMaxModifExtrema(d, MinEnv, MaxEnv, MinMat, MaxMat, data, ScaleIntensity, KSize, 0) ;
                        ////////////////////////////////////////// Compute the mean envelope
                        DataMat Mean(nR, nC) ; 
                        ComputeMeanEnvelope(d, MinEnv, MaxEnv, Mean) ; 

                        ////////////////////////////////////////// Reset maxima, except at borders
                        for (int i(0); i<nR; i++)
                                for(int j(0); j<nC; j++)	 
                                        {MinMat(i,j) = 0 ; MaxMat(i,j) = 0 ;}
                        d = Mean ; // repeat the process on the mean nSifts number of times for each mode
                }
		/////////////////////////////////////////////////////////////////////////////////////////////// End of iteration to find mean
                // Increase window for extrema detection for the succeeding mode
                KSize = 2*KSize - 1  ;
                // Compute detail by subtracting the mean 'd' from the previous residual 'd2'
                // Initially d2 = input image. After each smoothing operation d2 is the smoothed mean
                DataMat Detail(d2) ; 
                Detail-=d; 
                d2=d;
                _IMFs.push_back(Detail) ;
                nm = l ;
        } 
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////////////////// End of iteration over modes 
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        _IMFs.push_back(d) ;
}

